#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# See https://wiki.debian.org/UpstreamMetadata and
# https://dep-team.pages.debian.net/deps/dep12/

ADDON_ONLY_FIELDS = ["Archive"]
DEP12_FIELDS = [
    'Name',
    'Contact',
    'Archive',
    'ASCL-Id',
    'Bug-Database',
    'Bug-Submit',
    'Cite-As',
    'Changelog',
    'CPE',
    'Documentation',
    'Donation',
    'FAQ',
    'Funding',
    'Gallery',
    'Other-References',
    'Reference',
    'Registration',
    'Registry',
    'Repository',
    'Repository-Browse',
    'Screenshots',
    'Security-Contact',
    'Webservice',
]
